@@ -1,5 +1,5 @@
+import functools
 import inspect
-import types
 from dataclasses import dataclass
 from functools import lru_cache
 from typing import (
@@ -51,6 +51,8 @@ class DependencyParameter(Generic[T]):
 
 @lru_cache(maxsize=4096)
 def is_coroutine_callable(call: DependencyProvider) -> bool:
+    if isinstance(call, functools.partial):
+        call = call.func
     if inspect.isroutine(call):
         return inspect.iscoroutinefunction(call)
     if inspect.isclass(call):
@@ -78,7 +80,9 @@ def get_annotations(call: DependencyProvider) -> Dict[str, Any]:
     types_from: DependencyProvider
     if inspect.isclass(call):
         types_from = call.__init__  # type: ignore
-    elif not isinstance(call, types.FunctionType) and hasattr(call, "__call__"):
+    elif not (inspect.isfunction(call) or inspect.ismethod(call)) and hasattr(
+        call, "__call__"
+    ):
         # callable class
         types_from = call.__call__  # type: ignore
     else:
