@@ -1,115 +1,87 @@
+from __future__ import annotations
+
+import functools
 from collections import deque
-from contextlib import asynccontextmanager
 from contextvars import ContextVar
-from dataclasses import dataclass
 from typing import (
     Any,
-    AsyncGenerator,
+    Callable,
     ContextManager,
     Deque,
     Dict,
-    Generic,
     List,
     Optional,
     Tuple,
+    Union,
     cast,
 )
 
-from anyio import create_task_group
-from anyio.abc import TaskGroup
-
-from di._concurrency import bind_to_stack_as_awaitable, bind_to_stack_as_def_callable
 from di._inspect import (
     DependencyParameter,
     is_async_gen_callable,
     is_coroutine_callable,
 )
+from di._local_scope_context import LocalScopeContext
 from di._state import ContainerState
-from di._task import Task
+from di._task import AsyncTask, SyncTask, Task
 from di._topsort import topsort
-from di.dependency import (
-    DependantProtocol,
-    Dependency,
-    DependencyProvider,
-    DependencyProviderType,
-    DependencyType,
-    Scope,
-)
 from di.exceptions import (
     DependencyRegistryError,
     DuplicateScopeError,
     ScopeViolationError,
     UnknownScopeError,
 )
-
-_UNSET_SCOPE: Scope = object()
-
-
-@dataclass
-class SolvedDependency(Generic[DependencyType]):
-    """Representation of a fully solved dependency.
-
-    A fully solved dependency consists of:
-    - A DAG of sub-dependency paramters.
-    - A topologically sorted order of execution, where each sublist represents a
-    group of dependencies that can be executed in parallel.
-    """
-
-    dependency: DependantProtocol[DependencyType]
-    dag: Dict[
-        DependantProtocol[Any], Dict[str, DependencyParameter[DependantProtocol[Any]]]
-    ]
-    topsort: List[List[DependantProtocol[Any]]]
+from di.executors import DefaultExecutor
+from di.types import FusedContextManager
+from di.types.dependencies import DependantProtocol
+from di.types.executor import AsyncExecutor, SyncExecutor
+from di.types.providers import (
+    Dependency,
+    DependencyProvider,
+    DependencyProviderType,
+    DependencyType,
+)
+from di.types.scopes import Scope
+from di.types.solved import SolvedDependency
 
 
 class Container:
-    def __init__(self) -> None:
+    def __init__(
+        self, executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None
+    ) -> None:
         self._context = ContextVar[ContainerState]("context")
         state = ContainerState()
-        # bind ourself so that dependencies can request the container
+        state.cached_values.add_scope("container")
+        state.cached_values.set(Container, self, scope="container")
         self._context.set(state)
+        self._executor: Union[AsyncExecutor, SyncExecutor] = (
+            executor or DefaultExecutor()
+        )
 
     @property
     def _state(self) -> ContainerState:
         return self._context.get()
 
-    @asynccontextmanager
-    async def enter_global_scope(self, scope: Scope) -> AsyncGenerator[None, None]:
+    def enter_global_scope(self, scope: Scope) -> FusedContextManager[None]:
         """Enter a global scope that is shared amongst threads and coroutines.
 
         If you enter a global scope in one thread / coroutine, it will propagate to others.
         """
-        async with self._state.enter_scope(scope):
-            # bind ourself so that dependencies can request the container
-            if not self._state.cached_values.contains(Container):
-                self._state.cached_values.set(Container, self, scope=scope)
-            yield
+        return self._state.enter_scope(scope)
 
-    @asynccontextmanager
-    async def enter_local_scope(self, scope: Scope) -> AsyncGenerator[None, None]:
+    def enter_local_scope(self, scope: Scope) -> FusedContextManager[None]:
         """Enter a local scope that is localized to the current thread or coroutine.
 
         If you enter a global scope in one thread / coroutine, it will NOT propagate to others.
         """
         if scope in self._state.stacks:
             raise DuplicateScopeError(f"Scope {scope} has already been entered!")
-        current = self._state
-        new = current.copy()
-        token = self._context.set(new)
-        try:
-            async with new.enter_scope(scope):
-                # bind ourself so that dependencies can request the container
-                if not new.cached_values.contains(Container):
-                    new.cached_values.set(Container, self, scope=scope)
-                yield
-        finally:
-            self._context.reset(token)
+        return LocalScopeContext(self._context, scope)
 
     def bind(
         self,
         provider: DependantProtocol[DependencyType],
         dependency: DependencyProviderType[DependencyType],
-        scope: Optional[Scope] = _UNSET_SCOPE,
     ) -> ContextManager[None]:
         """Bind a new dependency provider for a given dependency.
 
@@ -123,13 +95,7 @@ class Container:
         The bind will be automatically cleared when that scope is exited.
         If no scope is provided, the current scope is used.
         """
-        if scope is _UNSET_SCOPE:
-            if len(self._state.scopes) == 0:
-                raise ValueError(
-                    "If there is no active scope, you MUST supply the `scope` parameter"
-                )
-            scope = self._state.scopes[-1]  # current scope
-        return self._state.bind(provider=provider, dependency=dependency, scope=scope)
+        return self._state.bind(provider=provider, dependency=dependency)
 
     def solve(
         self, dependency: DependantProtocol[DependencyType]
@@ -141,8 +107,8 @@ class Container:
         will not be changing between calls.
         """
 
-        if self._state.binds.contains(dependency.call):  # type: ignore
-            dependency = self._state.binds.get(dependency.call)  # type: ignore
+        if dependency.call in self._state.binds:
+            dependency = self._state.binds[dependency.call]  # type: ignore
 
         param_graph: Dict[
             DependantProtocol[Any],
@@ -159,7 +125,7 @@ class Container:
             params = dep.get_dependencies().copy()
             for keyword, param in params.items():
                 assert param.dependency.call is not None
-                if self._state.binds.contains(param.dependency.call):
+                if param.dependency.call in self._state.binds:
                     params[keyword] = DependencyParameter[Any](
                         dependency=self._state.binds.get(param.dependency.call),
                         parameter=param.parameter,
@@ -195,68 +161,67 @@ class Container:
                     if subdep not in dep_registry:
                         q.append(subdep)
 
-        groups = topsort(dependency, dep_dag)
-        return SolvedDependency[DependencyType](
-            dependency=dependency, dag=param_graph, topsort=groups
+        topsorted_groups = topsort(dependency, dep_dag)
+        tasks, get_results = self._build_tasks(
+            topsorted_groups, dependency, param_graph
+        )
+        return SolvedDependency(
+            dependency=dependency,
+            dag=param_graph,
+            _tasks=tasks,
+            _get_results=get_results,
         )
-
-    def get_flat_subdependants(
-        self, solved: SolvedDependency[Any]
-    ) -> List[DependantProtocol[Any]]:
-        """Get an exhaustive list of all of the dependencies of this dependency,
-        in no particular order.
-        """
-        return [dep for group in solved.topsort[1:] for dep in group]
 
     def _build_task(
         self,
         dependency: DependantProtocol[DependencyType],
-        tasks: Dict[DependantProtocol[Any], Tuple[DependantProtocol[Any], Task[Any]]],
-        state: ContainerState,
+        tasks: Dict[
+            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
+        ],
         dag: Dict[
             DependantProtocol[Any],
             Dict[str, DependencyParameter[DependantProtocol[Any]]],
         ],
-        parallel: bool,
-    ) -> Task[DependencyType]:
-        stack = state.stacks[dependency.scope]
-
-        async def bound_call(*args: Any, **kwargs: Any) -> DependencyType:
-            assert dependency.call is not None
-            if dependency.shared and state.cached_values.contains(dependency.call):
-                # use cached value
-                res = state.cached_values.get(dependency.call)
-            else:
-                # if this task is not being parallelized and we are dealing with a sync function
-                # then we can just execute it directly
-                # otherwise, we wrap it
-                if not parallel and not (
-                    is_coroutine_callable(dependency.call)
-                    or is_async_gen_callable(dependency.call)
-                ):
-                    res = bind_to_stack_as_def_callable(dependency.call, stack=stack)(
-                        *args, **kwargs
-                    )
-                else:
-                    res = await bind_to_stack_as_awaitable(
-                        dependency.call, stack=stack
-                    )(*args, **kwargs)
-                if dependency.shared:
-                    # caching is allowed, now that we have a value we can save it and start using the cache
-                    state.cached_values.set(
-                        dependency.call, res, scope=dependency.scope
-                    )
-
-            return cast(DependencyType, res)
+    ) -> Union[AsyncTask[DependencyType], SyncTask[DependencyType]]:
 
         task_dependencies: Dict[str, DependencyParameter[Task[DependencyProvider]]] = {}
 
         for param_name, param in dag[dependency].items():
             task_dependencies[param_name] = DependencyParameter(
-                dependency=tasks[param.dependency][1], parameter=param.parameter
+                dependency=tasks[param.dependency], parameter=param.parameter
             )
 
-        return Task(call=bound_call, dependencies=task_dependencies)
+        if is_async_gen_callable(dependency.call) or is_coroutine_callable(
+            dependency.call
+        ):
+            return AsyncTask(dependant=dependency, dependencies=task_dependencies)
+        else:
+            return SyncTask(dependant=dependency, dependencies=task_dependencies)
+
+    def _build_tasks(
+        self,
+        topsort: List[List[DependantProtocol[Any]]],
+        dependency: DependantProtocol[DependencyType],
+        dag: Dict[
+            DependantProtocol[Any],
+            Dict[str, DependencyParameter[DependantProtocol[Any]]],
+        ],
+    ) -> Tuple[
+        List[List[Union[AsyncTask[Dependency], SyncTask[Dependency]]]],
+        Callable[[], DependencyType],
+    ]:
+        tasks: Dict[
+            DependantProtocol[Any], Union[AsyncTask[Dependency], SyncTask[Dependency]]
+        ] = {}
+        for group in reversed(topsort):
+            for dep in group:
+                if dep not in tasks:
+                    tasks[dep] = self._build_task(dep, tasks, dag)
+        get_result = tasks[dependency].get_result
+        return (
+            list(reversed([[tasks[dep] for dep in group] for group in topsort])),
+            get_result,
+        )
 
     def _validate_scopes(self, solved: SolvedDependency[Dependency]) -> None:
         """Validate that dependencies all have a valid scope and
@@ -290,49 +255,60 @@ class Container:
                 check_scope(subdep)
                 check_is_inner(dep, subdep)
 
-    async def execute_solved(
-        self, solved: SolvedDependency[DependencyType], validate_scopes: bool = True
+    def execute_sync(
+        self,
+        solved: SolvedDependency[DependencyType],
+        validate_scopes: bool = True,
+    ) -> DependencyType:
+        """Execute an already solved dependency.
+
+        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
+        and then disable scope validation in subsequent runs with `validate_scope=False`.
+        """
+        with self.enter_local_scope(None):
+            if validate_scopes:
+                self._validate_scopes(solved)
+
+            tasks, get_results = solved._tasks, solved._get_results  # type: ignore
+
+            if not hasattr(self._executor, "execute_sync"):
+                raise TypeError(
+                    "execute_sync requires an executor implementing the SyncExecutor protocol"
+                )
+            executor = cast(SyncExecutor, self._executor)
+
+            stateful_tasks = [
+                [functools.partial(t.compute, self._state) for t in group]
+                for group in tasks
+            ]
+
+            return executor.execute_sync(stateful_tasks, get_results)  # type: ignore
+
+    async def execute_async(
+        self,
+        solved: SolvedDependency[DependencyType],
+        validate_scopes: bool = True,
     ) -> DependencyType:
         """Execute an already solved dependency.
 
         If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
         and then disable scope validation in subsequent runs with `validate_scope=False`.
         """
-        # this mapping uses the hash semantics defined by the implementation of DependantProtocol
-        tasks: Dict[
-            DependantProtocol[Any], Tuple[DependantProtocol[Any], Task[Any]]
-        ] = {}
         async with self.enter_local_scope(None):
             if validate_scopes:
                 self._validate_scopes(solved)
-            for group in reversed(solved.topsort):
-                parallel = len(group) > 1
-                for dep in group:
-                    if dep not in tasks:
-                        tasks[dep] = (
-                            dep,
-                            self._build_task(
-                                dep, tasks, self._state, solved.dag, parallel
-                            ),
-                        )
-            ordered_tasks = [
-                [tasks[dep][1] for dep in group] for group in solved.topsort
+
+            tasks, get_results = solved._tasks, solved._get_results  # type: ignore
+
+            if not hasattr(self._executor, "execute_async"):
+                raise TypeError(
+                    "execute_async requires an executor implementing the AsyncExecutor protocol"
+                )
+            executor = cast(AsyncExecutor, self._executor)
+
+            stateful_tasks = [
+                [functools.partial(t.compute, self._state) for t in group]
+                for group in tasks
             ]
-            tg: Optional[TaskGroup] = None
-            for task_group in reversed(ordered_tasks):
-                if len(task_group) > 1:
-                    if tg is None:
-                        tg = create_task_group()
-                    async with tg:
-                        for task in task_group:
-                            tg.start_soon(task.compute)  # type: ignore
-                else:
-                    await task_group[0].compute()
 
-        return tasks[solved.dependency][1].get_result()
-
-    async def execute(
-        self, dependency: DependantProtocol[DependencyType]
-    ) -> DependencyType:
-        """Solve and execute a dependency"""
-        return await self.execute_solved(self.solve(dependency))
+            return await executor.execute_async(stateful_tasks, get_results)  # type: ignore
